﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows.Forms;
using HeuristicLab.Collections;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.RealVectorEncoding;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis;

namespace HeuristicLab.ProcessParameterOptimization {
  [Item("Process parameter optimization problem (single-objective)", "Represents a single objective optimization problem which uses configurable regression solutions to evaluate targets from a given dataset.")]
  [Creatable("Problems")]
  [StorableClass]
  public sealed class SingleObjectiveGoalSeekingProblem : SingleObjectiveBasicProblem<RealVectorEncoding>, IGoalSeekingProblem {
    private const string ModifiableDatasetParameterName = "Dataset";
    private const string ProblemDataParameterName = "ProblemData";
    private const string ControllableParametersParameterName = "ControllableParameters";
    private const string ControllableParameterBoundsParameterName = "ControllableParameterBounds";
    private const string TargetGoalsParameterName = "TargetGoals";
    private const string TargetsParameterName = "Targets";
    private const string ModelCollectionParameterName = "ModelCollection";
    private const string RowParameterName = "Row";
    // these parameters are used by the pareto folding analyzer
    private const string AllowedRangesParameterName = "AllowedRanges";
    private const string QualitySumCutoffParameterName = "QualitySumCutoff";

    #region parameters
    public IValueParameter<RealVector> AllowedRangesParameter {
      get { return (IValueParameter<RealVector>)Parameters[AllowedRangesParameterName]; }
    }
    public IFixedValueParameter<DoubleValue> QualitySumCutoffParameter {
      get { return (IFixedValueParameter<DoubleValue>)Parameters[QualitySumCutoffParameterName]; }
    }
    public IValueParameter<IRegressionProblemData> ProblemDataParameter {
      get { return (IValueParameter<IRegressionProblemData>)Parameters[ProblemDataParameterName]; }
    }
    public IValueParameter<IDataset> ModifiableDatasetParameter {
      get { return (IValueParameter<IDataset>)Parameters[ModifiableDatasetParameterName]; }
    }
    public IValueParameter<CheckedItemList<StringValue>> ControllableParametersParameter {
      get { return (IValueParameter<CheckedItemList<StringValue>>)Parameters[ControllableParametersParameterName]; }
    }
    public IValueParameter<DoubleMatrix> ControllableParameterBoundsParameter {
      get { return (IValueParameter<DoubleMatrix>)Parameters[ControllableParameterBoundsParameterName]; }
    }
    public IFixedValueParameter<ItemCollection<IRegressionModel>> ModelCollectionParameter {
      get { return (IFixedValueParameter<ItemCollection<IRegressionModel>>)Parameters[ModelCollectionParameterName]; }
    }
    public IValueParameter<CheckedItemList<StringValue>> TargetsParameter {
      get { return (IValueParameter<CheckedItemList<StringValue>>)Parameters[TargetsParameterName]; }
    }
    public IValueParameter<DoubleMatrix> TargetGoalsParameter {
      get { return (IValueParameter<DoubleMatrix>)Parameters[TargetGoalsParameterName]; }
    }
    public IFixedValueParameter<IntValue> RowParameter {
      get { return (IFixedValueParameter<IntValue>)Parameters[RowParameterName]; }
    }
    #endregion

    #region parameter properties
    private IItemCollection<IRegressionModel> ModelCollection {
      get { return ModelCollectionParameter.Value; }
    }
    public DoubleMatrix TargetGoals {
      get { return TargetGoalsParameter.Value; }
      set { TargetGoalsParameter.Value = value; }
    }
    // the AllowedRanges property is only used by the pareto folding analyzer
    public RealVector AllowedRanges {
      get { return AllowedRangesParameter.Value; }
      set { AllowedRangesParameter.Value = value; }
    }
    public double QualitySumCutoff {
      get { return QualitySumCutoffParameter.Value.Value; }
      set { QualitySumCutoffParameter.Value.Value = value; }
    }
    #endregion

    #region IProcessParameterOptimizationProblem properties
    [Storable]
    private IRegressionProblemData problemData;
    public IRegressionProblemData ProblemData {
      get { return problemData; }
      set {
        if (value == null || value == problemData) return;
        var variables = value.Dataset.DoubleVariables.ToList();
        if (Models.Any()) {
          var targets = Models.Select(x => x.TargetVariable);
          var hashset = new HashSet<string>(variables);
          foreach (var target in targets) {
            if (!hashset.Contains(target)) {
              throw new ArgumentException(string.Format("Incompatible problem data. Target \"{0}\" is missing.", target));
            }
          }
        }
        problemData = value;
        dataset = new ModifiableDataset(variables, variables.Select(x => new List<double> { ProblemData.Dataset.GetDoubleValue(x, Row) }));
        ProblemDataParameter.Value = ProblemData;
        UpdateControllableParameters();
        UpdateTargetList();
      }
    }

    public int Row {
      get { return RowParameter.Value.Value; }
      set { RowParameter.Value.Value = value; }
    }

    public IEnumerable<IRegressionModel> Models {
      get { return ModelCollectionParameter.Value; }
    }

    #region targets
    public IEnumerable<string> ActiveTargets {
      get { return TargetList.CheckedItems.Select(x => x.Value.Value); }
    }

    public IEnumerable<string> Targets {
      get { return TargetList.Select(x => x.Value); }
    }

    public ICheckedItemList<StringValue> TargetList {
      get { return TargetsParameter.Value; }
      set { TargetsParameter.Value = (CheckedItemList<StringValue>)value; }
    }
    #endregion

    #region parameters
    public IEnumerable<string> ActiveParameters {
      get { return ControllableParameters.CheckedItems.Select(x => x.Value.Value); }
    }

    public ICheckedItemList<StringValue> ControllableParameters {
      get { return ControllableParametersParameter.Value; }
      set { ControllableParametersParameter.Value = (CheckedItemList<StringValue>)value; }
    }

    public DoubleMatrix ControllableParameterBounds {
      get { return ControllableParameterBoundsParameter.Value; }
      set { ControllableParameterBoundsParameter.Value = value; }
    }
    #endregion
    #endregion

    #region IProcessParameterOptimizationProblem methods
    #region solutions
    public double GetVariableValue(string variableName) {
      return dataset.GetDoubleValue(variableName, 0);
    }

    public void SetVariableValue(string variableName, double variableValue) {
      dataset.SetVariableValue(variableValue, variableName, 0); // will throw if the variable is not present in the dataset
    }

    public double[] GetEstimatedGoalValues(IEnumerable<double> parameters, bool round = false) {
      var ds = (ModifiableDataset)dataset.Clone();
      var estimatedValues = new double[ActiveTargets.Count()];
      var e1 = parameters.GetEnumerator();
      var e2 = ActiveParameters.GetEnumerator();
      while (e1.MoveNext() && e2.MoveNext())
        ds.SetVariableValue(e1.Current, e2.Current, 0);
      var rows = new[] { 0 }; // actually just one row
      int i = 0;
      var models = ModelCollection.ToList();
      foreach (var target in TargetList.CheckedItems) {
        var average = 0d;
        var count = 0;
        var targetName = target.Value.Value;

        foreach (var model in models) {
          if (targetName != model.TargetVariable)
            continue;
          average += model.GetEstimatedValues(ds, rows).Single();
          count++;
        }
        var estimatedValue = average / count;
        if (round) {
          var accuracy = GetTargetStepSize(targetName);
          estimatedValues[i] = RoundToNearestStepMultiple(estimatedValue, accuracy);
        } else {
          estimatedValues[i] = estimatedValue;
        }
        ++i;
      }
      return estimatedValues;
    }

    public void AddModel(IRegressionModel model) {
      var target = model.TargetVariable;
      CheckIfDatasetContainsTarget(target);
      ModelCollection.Add(model);
      OnModelsChanged(this, EventArgs.Empty);
    }

    // method which throws an exception that can be caught in the event handler if the check fails
    private void CheckIfDatasetContainsTarget(string target) {
      if (dataset.DoubleVariables.All(x => x != target))
        throw new ArgumentException(string.Format("Model target \"{0}\" does not exist in the dataset.", target));
    }

    public void RemoveModel(IRegressionModel model) {
      ModelCollection.Remove(model);
      OnModelsChanged(this, EventArgs.Empty);
    }

    public event EventHandler ModelsChanged;
    private void OnModelsChanged(object sender, EventArgs args) {
      var changed = ModelsChanged;
      if (changed == null) return;
      changed(sender, args);
    }
    #endregion

    #region targets
    public bool GetTargetActive(string target) {
      var item = TargetList.SingleOrDefault(x => x.Value == target);
      if (item == null)
        throw new ArgumentException(string.Format("SetTargetActive: Invalid target name {0}", target));
      return TargetList.ItemChecked(item);
    }

    public void SetTargetActive(string target, bool active) {
      var item = TargetList.SingleOrDefault(x => x.Value == target);
      if (item == null)
        throw new ArgumentException(string.Format("SetTargetActive: Invalid target name {0}", target));
      TargetList.SetItemCheckedState(item, active);
      OnTargetsChanged(this, EventArgs.Empty);
    }

    public double GetTargetGoal(string target) {
      if (!Targets.Contains(target))
        throw new ArgumentException(string.Format("The target variable name \"{0}\" does not exist in the dataset.", target));
      int i = TargetGoals.RowNames.TakeWhile(x => x != target).Count();
      return TargetGoals[i, 0];
    }

    public void SetTargetGoal(string target, double goal) {
      if (!Targets.Contains(target))
        throw new ArgumentException(string.Format("The target variable name \"{0}\" does not exist in the dataset.", target));
      int i = TargetGoals.RowNames.TakeWhile(x => x != target).Count();
      TargetGoals[i, 0] = goal;
      OnTargetsChanged(this, EventArgs.Empty);
    }

    public double GetTargetWeight(string target) {
      if (!Targets.Contains(target))
        throw new ArgumentException(string.Format("The target variable name \"{0}\" does not exist in the dataset.", target));
      int i = TargetGoals.RowNames.TakeWhile(x => x != target).Count();
      return TargetGoals[i, 1];
    }

    public void SetTargetWeight(string target, double weight) {
      if (!Targets.Contains(target))
        throw new ArgumentException(string.Format("The target variable name \"{0}\" does not exist in the dataset.", target));
      int i = TargetGoals.RowNames.TakeWhile(x => x != target).Count();
      TargetGoals[i, 1] = weight;
      OnTargetsChanged(this, EventArgs.Empty);
    }

    public double GetTargetVariance(string target) {
      if (!Targets.Contains(target))
        throw new ArgumentException(string.Format("The target variable name \"{0}\" does not exist in the dataset.", target));
      int i = TargetGoals.RowNames.TakeWhile(x => x != target).Count();
      return TargetGoals[i, 2];
    }

    public void SetTargetVariance(string target, double variance) {
      if (!Targets.Contains(target))
        throw new ArgumentException(string.Format("The target variable name \"{0}\" does not exist in the dataset.", target));
      int i = TargetGoals.RowNames.TakeWhile(x => x != target).Count();
      TargetGoals[i, 2] = variance;
      OnTargetsChanged(this, EventArgs.Empty);
    }

    public double GetTargetStepSize(string target) {
      if (!Targets.Contains(target))
        throw new ArgumentException(string.Format("The target variable name \"{0}\" does not exist in the dataset.", target));
      int i = TargetGoals.RowNames.TakeWhile(x => x != target).Count();
      return TargetGoals[i, 3];
    }

    public void SetTargetStepSize(string target, double stepSize) {
      if (!Targets.Contains(target))
        throw new ArgumentException(string.Format("The target variable name \"{0}\" does not exist in the dataset.", target));
      int i = TargetGoals.RowNames.TakeWhile(x => x != target).Count();
      TargetGoals[i, 3] = stepSize;
      OnTargetsChanged(this, EventArgs.Empty);
    }

    public event EventHandler TargetsChanged;
    private void OnTargetsChanged(object sender, EventArgs args) {
      var changed = TargetsChanged;
      if (changed == null) return;
      changed(sender, args);
    }
    #endregion // targets

    #region process parameters
    public void SetParameterBounds(string parameterName, double min, double max, double step) {
      int i = ControllableParameterBounds.RowNames.TakeWhile(x => x != parameterName).Count();
      if (i < ControllableParameterBounds.Rows) {
        ControllableParameterBounds[i, 0] = min;
        ControllableParameterBounds[i, 1] = max;
        ControllableParameterBounds[i, 2] = step;
        UpdateEncoding();
        OnParametersChanged(this, EventArgs.Empty);
      } else {
        throw new ArgumentException(string.Format("SetParameterBounds: Invalid parameter name {0}", parameterName));
      }
    }

    public double GetParameterStepSize(string parameter) {
      int i = ControllableParameterBounds.RowNames.TakeWhile(x => x != parameter).Count();
      if (i < ControllableParameterBounds.Rows)
        return ControllableParameterBounds[i, 2];
      throw new ArgumentException(string.Format("GetParameterStepSize: Invalid parameter name {0}", parameter));
    }

    public void SetParameterStepSize(string parameter, double stepSize) {
      int i = ControllableParameterBounds.RowNames.TakeWhile(x => x != parameter).Count();
      if (i < ControllableParameterBounds.Rows) {
        ControllableParameterBounds[i, 2] = stepSize;
        OnParametersChanged(this, EventArgs.Empty);
      } else {
        throw new ArgumentException(string.Format("SetParameterStepSize: Invalid parameter name {0}", parameter));
      }
    }

    public bool GetParameterActive(string parameter) {
      var item = ControllableParameters.SingleOrDefault(x => x.Value == parameter);
      if (item == null)
        throw new ArgumentException(string.Format("GetParameterActive: Invalid target name {0}", parameter));
      return ControllableParameters.ItemChecked(item);
    }

    public void SetParameterActive(string parameter, bool active) {
      var item = ControllableParameters.SingleOrDefault(x => x.Value == parameter);
      if (item == null)
        throw new ArgumentException(string.Format("SetParameterActive: Invalid target name {0}", parameter));
      ControllableParameters.SetItemCheckedState(item, active);
      OnParametersChanged(this, EventArgs.Empty);
    }

    /// <summary>
    /// Returns the parameter bounds (min and max) and the step size for the specified parameter
    /// </summary>
    /// <param name="parameterName"></param>
    /// <returns>A double array containing the values (min, max, step) in this order</returns>
    public double[] GetParameterBounds(string parameterName) {
      var index = ControllableParameters.TakeWhile(x => x.Value != parameterName).Count();
      if (index < ControllableParameters.Count) {
        var min = ControllableParameterBounds[index, 0];
        var max = ControllableParameterBounds[index, 1];
        var step = ControllableParameterBounds[index, 2];
        return new[] { min, max, step };
      }
      throw new ArgumentException(string.Format("GetParameterBounds: Unknown parameter {0}.", parameterName));
    }

    public void SetControllableParameters(IEnumerable<string> parameterNames) {
      ControllableParameters = new CheckedItemList<StringValue>();
      foreach (var v in parameterNames) {
        ControllableParameters.Add(new StringValue(v), false);
      }
      ControllableParameters.CheckedItemsChanged += ControllableParameters_OnItemsChanged;
      ControllableParameterBounds = new DoubleMatrix(ControllableParameters.Count, 3);
      ControllableParameterBounds.RowNames = GetControllableParameters();
      ControllableParameterBounds.ColumnNames = new[] { "Min", "Max", "Step" };

      for (int i = 0; i < ControllableParameters.Count; ++i) {
        var itemName = ControllableParameters[i].Value;
        var values = ProblemData.Dataset.GetReadOnlyDoubleValues(itemName).Where(x => !double.IsNaN(x) && !double.IsInfinity(x)).ToList();
        if (!values.Any()) continue;

        // add a 20% margin to allow the optimization algorithm more freedom of exploration
        ControllableParameterBounds[i, 0] = 0.8 * values.Min(); // min
        ControllableParameterBounds[i, 1] = 1.2 * values.Max(); // max
        ControllableParameterBounds[i, 2] = 1e-6;               // step
      }
      OnParametersChanged(this, EventArgs.Empty);
    }

    public IEnumerable<string> GetControllableParameters() {
      return ControllableParameters.Select(x => x.Value);
    }

    public event EventHandler ParametersChanged;
    private void OnParametersChanged(object sender, EventArgs args) {
      var changed = ParametersChanged;
      if (changed == null) return;
      changed(sender, args);
    }
    #endregion // process parameters
    #endregion // IProcessParameterOprimizationProblem methods

    #region data members
    [Storable]
    private ModifiableDataset dataset; // modifiable dataset

    public override bool Maximization {
      get { return true; }
    }
    #endregion

    #region constructors
    [StorableConstructor]
    private SingleObjectiveGoalSeekingProblem(bool deserializing) : base(deserializing) { }

    private SingleObjectiveGoalSeekingProblem(SingleObjectiveGoalSeekingProblem original, Cloner cloner)
      : base(original, cloner) {
      this.dataset = cloner.Clone(original.dataset);
      this.problemData = cloner.Clone(original.problemData);

      RegisterEvents();
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new SingleObjectiveGoalSeekingProblem(this, cloner);
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      // backwards-compatibility
      if (Parameters.ContainsKey("Models")) {
        var solutions = ((IFixedValueParameter<ItemCollection<IRegressionSolution>>)Parameters["Models"]).Value;
        var models = new ItemCollection<IRegressionModel>();
        foreach (var solution in solutions) {
          var model = solution.Model;
          model.TargetVariable = solution.ProblemData.TargetVariable;
          models.Add(model);
        }
        if (Parameters.ContainsKey(ModelCollectionParameterName))
          Parameters.Remove(ModelCollectionParameterName);
        Parameters.Add(new FixedValueParameter<ItemCollection<IRegressionModel>>(ModelCollectionParameterName, models));
      }

      RegisterEvents();
    }

    public SingleObjectiveGoalSeekingProblem() {
      Parameters.Add(new ValueParameter<IRegressionProblemData>(ProblemDataParameterName, new RegressionProblemData()));
      Parameters.Add(new ValueParameter<IDataset>(ModifiableDatasetParameterName, dataset) { Hidden = true });
      Parameters.Add(new ValueParameter<CheckedItemList<StringValue>>(ControllableParametersParameterName));
      Parameters.Add(new ValueParameter<CheckedItemList<StringValue>>(TargetsParameterName));
      Parameters.Add(new ValueParameter<DoubleMatrix>(ControllableParameterBoundsParameterName));
      Parameters.Add(new FixedValueParameter<ItemCollection<IRegressionModel>>(ModelCollectionParameterName, new ItemCollection<IRegressionModel>()));
      Parameters.Add(new ValueParameter<DoubleMatrix>(TargetGoalsParameterName)); // model target weights
      Parameters.Add(new FixedValueParameter<IntValue>(RowParameterName));

      // when the problem is created, the problem data parameter will be set to a default value
      // set the internal property to the same value
      ProblemData = ProblemDataParameter.Value;

      UpdateControllableParameters();
      UpdateTargetList();
      RegisterEvents();
    }
    #endregion

    public override double Evaluate(Individual individual, IRandom random) {
      var vector = individual.RealVector();
      vector.ElementNames = ActiveParameters;

      int i = 0;
      // round vector according to parameter step sizes
      foreach (var parameter in ControllableParameters.CheckedItems) {
        var step = ControllableParameterBounds[parameter.Index, 2];
        vector[i] = RoundToNearestStepMultiple(vector[i], step);
        ++i;
      }
      i = 0;
      var quality = 0d;
      var estimatedValues = GetEstimatedGoalValues(vector, round: true);
      // round target estimated values and calculate qualities
      foreach (var target in ActiveTargets) {
        var estimatedValue = estimatedValues[i];
        var goal = GetTargetGoal(target);
        var weight = GetTargetWeight(target);
        var variance = GetTargetVariance(target);
        quality += weight * Math.Pow(estimatedValue - goal, 2) / variance;
        ++i;
      }
      return quality / i;
    }

    #region event handlers
    private void RegisterEvents() {
      ProblemDataParameter.ValueChanged += OnProblemDataChanged;
      ModelCollectionParameter.Value.ItemsAdded += ModelCollection_OnItemsAdded;
      ModelCollectionParameter.Value.ItemsRemoved += ModelCollection_OnItemsRemoved;
      RowParameter.Value.ValueChanged += OnRowChanged;
      ControllableParameters.CheckedItemsChanged += ControllableParameters_OnItemsChanged;
      ControllableParameterBounds.ItemChanged += ControllableParameterBounds_ItemChanged;
    }

    private void OnRowChanged(object o, EventArgs e) {
      // set variables in the modifiable dataset according to the new row
      foreach (var v in dataset.DoubleVariables)
        dataset.SetVariableValue(ProblemData.Dataset.GetDoubleValue(v, Row), v, 0);
      // set the correct targets
      UpdateTargetList();
    }

    private void OnProblemDataChanged(object o, EventArgs e) {
      try {
        ProblemData = ProblemDataParameter.Value;
      }
      catch (ArgumentException exception) {
        MessageBox.Show(exception.Message, "Update Problem Data", MessageBoxButtons.OK, MessageBoxIcon.Error);
        ProblemDataParameter.Value = problemData;
      }
    }

    private void ModelCollection_OnItemsAdded(object sender, CollectionItemsChangedEventArgs<IRegressionModel> e) {
      if (e.Items == null) return;

      var collection = (IObservableCollection<IRegressionModel>)sender;
      var newItems = e.Items.ToList();

      foreach (var model in e.Items) {
        try {
          CheckIfDatasetContainsTarget(model.TargetVariable);
        }
        catch (ArgumentException exception) {
          MessageBox.Show(exception.Message, "Add Model", MessageBoxButtons.OK, MessageBoxIcon.Error);
          newItems.Remove(model);
          collection.Remove(model);
        }
      }
      UpdateTargetList();
      OnModelsChanged(this, EventArgs.Empty);
    }

    private void ModelCollection_OnItemsRemoved(object sender, CollectionItemsChangedEventArgs<IRegressionModel> e) {
      if (e.Items == null) return;
      UpdateTargetList();
      OnModelsChanged(this, EventArgs.Empty);
    }

    private void ControllableParameters_OnItemsChanged(object o, CollectionItemsChangedEventArgs<IndexedItem<StringValue>> e) {
      UpdateEncoding();
    }

    private void ControllableParameterBounds_ItemChanged(object o, EventArgs e) {
      UpdateEncoding();
    }
    #endregion

    #region helper methods
    private void UpdateControllableParameters() {
      if (ProblemData == null) return;
      SetControllableParameters(ProblemData.Dataset.DoubleVariables);
    }

    private void UpdateTargetList() {
      if (ProblemData == null) return;
      if (!Models.Any()) {
        TargetGoals = new DoubleMatrix();
        TargetList = new CheckedItemList<StringValue>();
        return;
      }
      var targetNames = Models.Select(x => x.TargetVariable).Distinct().ToList();
      var oldTargetGoals = (DoubleMatrix)TargetGoals.Clone();
      var oldRowIndices = oldTargetGoals.RowNames.Select((x, i) => new { x, i }).ToDictionary(x => x.x, x => x.i);
      TargetGoals = new DoubleMatrix(targetNames.Count, 4);
      TargetGoals.RowNames = targetNames;
      TargetGoals.ColumnNames = new[] { "Goal", "Weight", "Variance", "Step size" };

      TargetList = new CheckedItemList<StringValue>();
      for (int i = 0; i < targetNames.Count; ++i) {
        TargetList.Add(new StringValue(targetNames[i]), true);
        int rowIndex;
        if (oldRowIndices.TryGetValue(targetNames[i], out rowIndex)) {
          for (int j = 0; j < TargetGoals.Columns; ++j)
            TargetGoals[i, j] = oldTargetGoals[rowIndex, j];
        } else {
          TargetGoals[i, 0] = ProblemData.Dataset.GetDoubleValue(targetNames[i], Row);
          TargetGoals[i, 1] = 1.0;
          TargetGoals[i, 2] = ProblemData.Dataset.GetReadOnlyDoubleValues(targetNames[i]).Variance();
          TargetGoals[i, 3] = 1e-6;
        }
      }
    }

    private void UpdateEncoding() {
      var activeParameters = ActiveParameters.ToList();
      if (Encoding == null)
        Encoding = new RealVectorEncoding(activeParameters.Count);
      else
        Encoding.Length = activeParameters.Count;

      Encoding.Bounds = new DoubleMatrix(activeParameters.Count, 2); // only two columns: min and max
      Encoding.Bounds.RowNames = activeParameters;
      Encoding.Bounds.ColumnNames = new[] { "Min.", "Max." };

      int i = 0;
      foreach (var item in ControllableParameters.CheckedItems) {
        var index = item.Index;
        Encoding.Bounds[i, 0] = ControllableParameterBounds[index, 0];
        Encoding.Bounds[i, 1] = ControllableParameterBounds[index, 1];
        ++i;
      }
    }
    #endregion

    #region util
    private static double RoundToNearestStepMultiple(double value, double step) {
      return step * (long)Math.Round(value / step);
    }
    #endregion
  }
}
